(*
$ cd /usr/local/isabisac/src/Pure/isac/smltest/Pure/Isar
$ /usr/local/isabisac/bin/isabelle emacs Struct_Der.thy &
*)

header {* structured derivations (SD) according to R.J.Back *}

theory Struct_Der
imports Main
begin

section {* parsing SDs *}

subsection {* keywords, outer syntax and parser setup *}
text {* setup is still a copy from respective Isabelle sourcefiles,
        trying to touch the sources presumably to be adapted to SD.*}

subsubsection {* keywords according to src/Pure/Isar/keyword.ML*}
ML {*
signature KEYWORDC =
sig
  type T
  val kind_of: T -> string
  val calc_begin: T
  val keyword: string -> unit
  val command: string -> T -> unit
  val tag_calc: T -> T
end;

structure KeywordC: KEYWORDC =
struct

datatype T = KeywordC of string * string list;

fun kind s = KeywordC (s, []);
fun kind_of (KeywordC (s, _)) = s;

val calc_begin = kind "calc-begin";

(*
fun change_commands f = CRITICAL (fn () => Unsynchronized.change global_commands f);
fun change_lexicons f = CRITICAL (fn () => Unsynchronized.change global_lexicons f);
fun report_message s =
  (if print_mode_active keyword_status_reportN then Output.status else writeln) s;*)
fun report_message s =
  (if true (*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*) then Output.status else writeln) s;
fun report_keyword name =
  report_message (Markup.markup (Markup.keyword_decl name)
    ("Outer syntax keyword: " ^ quote name));
fun report_command (name, kind) =
  report_message (Markup.markup (Markup.command_decl name (kind_of kind))
    ("Outer syntax command: " ^ quote name ^ " (" ^ kind_of kind ^ ")"));

fun keyword name =
 (apfst (Scan.extend_lexicon (Symbol.explode name));
  report_keyword name);
(* see CRITICAL above
fun command name kind =
 (change_lexicons (apsnd (Scan.extend_lexicon (Symbol.explode name)));
  change_commands (Symtab.update (name, kind));
  report_command (name, kind));*)
fun command name kind =
 (apsnd (Scan.extend_lexicon (Symbol.explode name));
  Symtab.update (name, kind);
  report_command (name, kind));

fun update_tags t ts = t :: remove (op = : string * string -> bool) t ts;
fun tag t (KeywordC (s, ts)) = KeywordC (s, update_tags t ts);
val tag_calc = tag "calculation"; (* <----------------------------------------*)

end; *}

subsubsection {* outer syntax according to src/Pure/Isar/outer_syntax.ML*}
ML {* 
signature OUTER_SYNTAXC = 
sig
  val command: string -> string -> KeywordC.T ->
    (Toplevel.transition -> Toplevel.transition) parser -> unit
  val scan: Position.T -> string -> Token.T list
end;

structure Outer_SyntaxC: OUTER_SYNTAXC =
struct

datatype command = CommandC of
 {comment: string,
  markup: Thy_Output.markup option,
  int_only: bool,
  parse: (Toplevel.transition -> Toplevel.transition) parser};

fun make_command comment markup int_only parse =
  CommandC {comment = comment, markup = markup, int_only = int_only, parse = parse};

val global_commands = Unsynchronized.ref (Symtab.empty: command Symtab.table);
val global_markups = Unsynchronized.ref ([]: (string * Thy_Output.markup) list);

fun change_commands f = CRITICAL (fn () =>
 (Unsynchronized.change global_commands f;
  global_markups :=
    Symtab.fold (fn (name, CommandC {markup = SOME m, ...}) => cons (name, m) | _ => I)
      (! global_commands) []));
fun get_commands () = ! global_commands;
fun add_command name kind cmd = CRITICAL (fn () =>
 (KeywordC.command name kind;
  if not (Symtab.defined (get_commands ()) name) then ()
  else warning ("Redefining outer syntax command " ^ quote name);
  change_commands (Symtab.update (name, cmd))));

fun command name comment kind parse =
  add_command name kind (make_command comment NONE false parse);

fun scan pos str =
  Source.of_string str
  |> Symbol.source {do_recover = false}
  |> Token.source {do_recover = SOME false} Keyword.get_lexicons pos
  |> Source.exhaust;
end; *}

subsubsection {* parser setup according to src/Pure/Isar/outer_syntax.ML*}
ML {* (*src/Pure/Isar/isar_syn.ML*)
structure Isac_Syn: sig end =
struct

val _ = List.app(**?find fun app*) KeywordC.keyword
 [(*"(", ")", "+", "/", "[", "]", ",", "<", "<=", "=", "=>", !!!!!!!!!!!!!!!!!!*)
  "CAS", "Box", "bullet", "dots", "==", "equiv", "top"];

val _ =
  Outer_SyntaxC.command "calculation" "begin calc" (KeywordC.tag_calc KeywordC.calc_begin)
    (Thy_Header.args >> (Toplevel.print oo Isar_Cmd.init_theory)); (*TODO*)
end; *}

subsection {* build parser for SD *}

subsubsection {* fun parse *}
text {* see ~/devel/IsaDevWS/IsaDevWS-10/T05_Methods.thy by Stefan Berghofer *}
ML {*
fun filtered_input str =
  filter Token.is_proper (Outer_SyntaxC.scan Position.none str);
fun parse p str = Scan.finite Token.stopper p (filtered_input str);
*}

subsubsection {* examples *}

ML {*
val ex1 = "calculation                  \n\
  \ bullet CAS solve(x+1=2,x)           \n\
  \    top x + 1 = 2                    \n\
  \    equiv x + 1 - 2 = 0              \n\
  \    equiv -1 + x = 0                 \n\
  \    bullet CAS solve (-1 + x = 0, x) \n\
  \       top -1 + x = 0                \n\
  \       equiv x = -1 * -1             \n\
  \       equiv x = 1                   \n\
  \       equiv [x = 1]                 \n\
  \    dots [x = 1]                     \n\
  \ dots [x = 1]                        \n\
  \ Box ";

val ex2 = "calculation                                                                                                         \n\
  \ bullet Problem (Biegelinie.thy, [Biegelinien])                                                                             \n\
  \    bullet Problem (Biegelinie.thy, [vonBelastungZu, Biegelinien])                                                          \n\
  \    dots [Q x = c + -1 * q_0 * x, M_b x = c_2 + c * x + -1 * q_0 / 2 * x ^ 2,                                               \n\
  \          y' x = c_3 + 1 / (-1 * EI) * (c_2 * x + c / 2 * x ^ 2 + -1 * q_0 / 6 * x ^ 3),                                    \n\
  \          y x = c_4 + c_3 * x + 1 / (-1 * EI) * (c_2 / 2 * x ^ 2 + c / 6 * x ^ 3 + -1 * q_0 / 24 * x ^ 4)]                  \n\
  \    bullet Problem (Biegelinie.thy, [setzeRandbedingungen, Biegelinien])                                                    \n\
  \    dots [L * q_0 = c, 0 = (2 * c_2 + 2 * L * c + -1 * L ^ 2 * q_0) / 2, 0 = c_4, 0 = c_3]                                  \n\
  \    bullet CAS solveSystem [0 = c_3] [c_4]                                                                                  \n\
  \    dots [c = L * q_0, c_2 = -1 * L ^ 2 * q_0 / 2, c_3 = 0, c_4 = 0]                                                        \n\
  \    top y x = c_4 + c_3 * x + 1 / (-1 * EI) * (c_2 / 2 * x ^ 2 + c / 6 * x ^ 3 + -1 * q_0 / 24 * x ^ 4)                     \n\
  \    equiv y x = 0 + 0 * x + 1 / (-1 * EI) * (-1 * L ^ 2 * q_0 / 2 / 2 * x ^ 2 + L * q_0 / 6 * x ^ 3 + -1 * q_0 / 24 * x ^ 4)\n\
  \    equiv y x = -6 * q_0 * L ^ 2 / (-24 * EI) * x ^ 2 + 4 * L * q_0 / (-24 * EI) * x ^ 3 + -1 * q_0 / (-24 * EI) * x ^ 4    \n\
  \ dots y x = -6 * q_0 * L ^ 2 / (-24 * EI) * x ^ 2 + 4 * L * q_0 / (-24 * EI) * x ^ 3 + -1 * q_0 / (-24 * EI) * x ^ 4        \n\
  \ Box";
*}

subsubsection {* parsers for the "executable" elements *}

ML {*
val pbl = ["setzeRandbedingungen", "Biegelinien"];
commas pbl;
fun mkstr pbl = "[" ^ commas pbl ^ "]";
fn pbl => "[" ^ commas pbl ^ "]";
*}

ML {*
"---1----------------------------------------";
val headline = Args.$$$ "Problem" -- Args.$$$ "(" -- Args.name -- Args.$$$ ","
  -- (Parse.$$$ "[" |-- Parse.list Parse.name --| Parse.$$$ "]") 
  -- Args.$$$ ")";
parse headline "Problem (Biegelinie.thy, [Biegelinien])";
(*val it = (((((("Problem", "("), "Biegelinie.thy"), ","), ["Biegelinien"]), ")"), [])
   : (((((string * string) * string) * string) * bstring list) * string) * Token.T list*)

fun exec1 ((((("Problem", "("), thy             ), ","), pbl            ), ")") =
  "Problem (" ^ thy ^ ", " ^ ((fn ss => "[" ^ commas ss ^ "]") pbl) ^ ")"
  |> (fn x => (x, x))
  |>> writeln
  |> #2;
exec1     ((((("Problem", "("), "Biegelinie.thy"), ","), ["Biegelinien"]), ")");

"---2----------------------------------------";
parse (headline >> exec1) "Problem (Biegelinie.thy, [Biegelinien])";
parse (headline >> exec1) "Problem (Biegelinie.thy, [setzeRandbedingungen, Biegelinien])";
*}

text {* compare tokens in Parse.name and Parse.term *}
ML {*
parse (Scan.repeat Parse.term) "xxx +++ 111 === 222 end";
parse (Scan.repeat Parse.term) "(xxx +++ 111 === 222 end";

parse (Scan.repeat Parse.term) "xxx + 111 = 222 end";

*}

ML {*
val cas_headline = Args.$$$ "CAS" -- (Scan.repeat Parse.term);

parse cas_headline "CAS solve x+1=2, x";
*}

ML {* 
parse (Args.$$$ "bullet" -- 
  Args.$$$ "Problem" -- Args.$$$ "(" -- Args.name -- Args.$$$ "," 
  -- Args.$$$ "[" -- Args.name -- 
  Scan.repeat (Args.$$$ "," -- Args.name)) 
  "bullet Problem (Biegelinie.thy, [vonBelastungZu, Biegelinien])";

*}

subsubsection {* GOON *}
(*
ML {* *}
*)

end